package ch15_wordraetsel;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * Beispielprogramm für das Buch "Einfach Java"
 *
 * (C) 2021 Michael Inden
 */
public class Model
{
    private static final int MAX_TRYS         = 5;

    private final List<String>       allPossibleWords = new ArrayList<>();

    private final List<String>       wordsToSearch    = new ArrayList<>();

    private final char[][]           board            = new char[10][];

    
    
    Model(final List<String> wordsToUse, final int wordsToPlace)
    {
        allPossibleWords.addAll(wordsToUse);

        prepareBoard(wordsToPlace);
    }

    public void prepareBoard(final  int wordsToPlace)
    {
        initializeEmptyBoard();

        placeWords(wordsToPlace);

        fillUpWithRandomChars();
    }

    private void initializeEmptyBoard()
    {
        for (int y = 0; y < getBoardHeight(); y++)
        {
            board[y] = new char[getBoardWidth()];
            for (int x = 0; x < getBoardWidth(); x++)
            {
                board[y][x] = ' ';
            }
        }
    }

    public List<String> getWordsToSearch()
    {
        return Collections.unmodifiableList(wordsToSearch);
    }
    
    public char getAt(int x, int y)
    {
        return board[y][x];
    }
    
    public int getBoardWidth()
    {
        return 20;
    }

    public int getBoardHeight()
    {
        return 10;
    }
    
    private void placeWords(int wordsToPlace)
    {
        int impossibleToPlace = 0;
        while (wordsToPlace > 0)
        {
            if (placeSingleWord())
            {
                wordsToPlace--;               
            }
            else
            {
                // So verhindert man Endlosschleife
                impossibleToPlace++;
                if (impossibleToPlace > 5)
                    break;
            }
        }
    }

    boolean placeSingleWord()
    {
        boolean wordFits = false;

        final String word = chooseRandomWord();

        int currentTry = 0;
        while (currentTry < MAX_TRYS && !wordFits)
        {
            final Position pos = chooseRandomPosition();
            final Direction dir = chooseRandomDirection();

            wordFits = wordFitsOnBoard(word, pos, dir);
            if (wordFits)
            {
                putWordOnBoard(word, pos, dir);
                
                allPossibleWords.remove(word);
                wordsToSearch.add(word);
            }

            currentTry++;
        }

        return wordFits;
    }

    private static int randomRange(int max)
    {
        return (int) (max * Math.random());
        
        // Random r = new Random();
        // int rand = r.nextInt(50);
    }
    
    private static Direction chooseRandomDirection()
    {
        int dirIdex = randomRange(8);
        
        return Direction.values()[dirIdex];
    }

    private Position chooseRandomPosition()
    {
        final int randomX = randomRange(getBoardWidth());
        final int randomY = randomRange(getBoardHeight());

        return new Position(randomX, randomY);
    }

    private String chooseRandomWord()
    {
        final int wordIdx = randomRange(allPossibleWords.size());

        return allPossibleWords.get(wordIdx);
    }

    private char randomChar()
    {
        return (char) ('A' + randomRange(26));
    }   
    
    private boolean wordFitsOnBoard(String word, Position startPos, Direction dir)
    {
        final char[] chars = word.toCharArray();

        Position pos = startPos;

        for (int i = 0; i < chars.length; i++)
        {
            if (isOnBoard(pos) && isEmptyCell(pos))
            {
                pos = pos.adjustToDir(dir);
            }
            else
            {
                return false;
            }
        }

        return true;
    }

    private void putWordOnBoard(String word, Position startPos, Direction dir)
    {
        final char[] chars = word.toCharArray();

        Position pos = startPos;

        for (int i = 0; i < chars.length; i++)
        {
            board[pos.y()][pos.x()] = chars[i];
            pos = pos.adjustToDir(dir);
        }
    }
    
    private boolean isOnBoard(Position pos)
    {
        return (pos.y() >= 0 && pos.y() < getBoardHeight()) && 
               (pos.x() >= 0 && pos.x() < getBoardWidth());
    }

    private boolean isEmptyCell(Position pos)
    {
        return isEmptyCell(pos.x(), pos.y());
    }

    private boolean isEmptyCell(int x, int y)
    {
        return board[y][x] == ' ';
    }
    
    private void fillUpWithRandomChars()
    {
        for (int y = 0; y < getBoardHeight(); y++)
        {
            for (int x = 0; x < getBoardWidth(); x++)
            {
                if (isEmptyCell(x, y))
                {
                    board[y][x] = randomChar();
                }
            }
        }
    }
}
